@@ -1,8 +1,9 @@
-#' Find a few approximate largest singular values and corresponding
+#' Find a few approximate singular values and corresponding
 #' singular vectors of a matrix.
 #'
 #' The augmented implicitly restarted Lanczos bidiagonalization algorithm
-#' (IRLBA) finds a few approximate largest singular values and corresponding
+#' (IRLBA) finds a few approximate largest (or, optionally, smallest) singular
+#' values and corresponding
 #' singular vectors of a sparse or dense matrix using a method of Baglama and
 #' Reichel.  It is a fast and memory-efficient way to compute a partial SVD.
 #'
@@ -13,8 +14,11 @@
 #' @param work working subspace dimension, larger values can speed convergence at the cost of more memory use.
 #' @param reorth if \code{TRUE}, apply full reorthogonalization to both SVD bases, otherwise
 #'   only apply reorthogonalization to the right SVD basis vectors; the latter case is cheaper per
-#'   iteration but, overall, may require more iterations for convergence.
-#' @param tol convergence is determined when \eqn{\|AV - US\| < tol\|A\|}{||AV - US|| < tol*||A||},
+#'   iteration but, overall, may require more iterations for convergence. Automatically \code{TRUE}
+#'   when \code{fastpath=TRUE} (see below).
+#' @param tol convergence is determined when \eqn{\|A^TU - VS\| < tol\|A\|}{||A^T U - VS|| < tol*||A||},
+#'   and when the maximum relative change in estimated singular values from one iteration to the
+#'   next is less than \code{svtol = tol} (see \code{svtol} below),
 #'   where the spectral norm ||A|| is approximated by the
 #'   largest estimated singular value, and U, V, S are the matrices corresponding
 #'   to the estimated left and right singular vectors, and diagonal matrix of
@@ -23,19 +27,30 @@
 #'   to restart the algorithm from where it left off (see the notes).
 #' @param right_only logical value indicating return only the right singular vectors
 #'  (\code{TRUE}) or both sets of vectors (\code{FALSE}). The right_only option can be
-#'  cheaper to compute and use much less memory when \code{nrow(A) >> ncol(A)}.
+#'  cheaper to compute and use much less memory when \code{nrow(A) >> ncol(A)} but note
+#'  that \code{right_only = TRUE} sets \code{fastpath = FALSE} (only use this option
+#'  for really large problems that run out of memory and when \code{nrow(A) >> ncol(A)}).
 #' @param verbose logical value that when \code{TRUE} prints status messages during the computation.
 #' @param scale optional column scaling vector whose values divide each column of \code{A};
 #'   must be as long as the number of columns of \code{A} (see notes).
 #' @param center optional column centering vector whose values are subtracted from each
 #'   column of \code{A}; must be as long as the number of columns of \code{A} and may
-#"   not be used together with the deflation options below (see notes).
-#' @param du DEPRECATED optional subspace deflation vector (see notes).
-#' @param ds DEPRECATED optional subspace deflation scalar (see notes).
-#' @param dv DEPRECATED optional subspace deflation vector (see notes).
+#'   not be used together with the deflation options below (see notes).
 #' @param shift optional shift value (square matrices only, see notes).
-#' @param mult optional custom matrix multiplication function (default is \code{\%*\%}, see notes).
-#' @param fastpath try a fast C algorithm implementation if possible; set \code{fastpath=FALSE} to use the reference R implementation.
+#' @param mult DEPRECATED optional custom matrix multiplication function (default is \code{\%*\%}, see notes).
+#' @param fastpath try a fast C algorithm implementation if possible; set \code{fastpath=FALSE} to use the
+#'     reference R implementation. See the notes for more details.
+#' @param svtol additional stopping tolerance on maximum allowed absolute relative change across each
+#' estimated singular value between iterations.
+#' The default value of this parameter is to set it to \code{tol}. You can set \code{svtol=Inf} to
+#' effectively disable this stopping criterion. Setting \code{svtol=Inf} allows the method to
+#' terminate on the first Lanczos iteration if it finds an invariant subspace, but with less certainty
+#' that the converged subspace is the desired one. (It may, for instance, miss some of the largest
+#' singular values in difficult problems.)
+#' @param smallest set \code{smallest=TRUE} to estimate the smallest singular values and associated
+#' singular vectors. WARNING: this option is somewhat experimental, and may produce poor
+#' estimates for ill-conditioned matrices.
+#' @param ... optional additional arguments used to support experimental and deprecated features.
 #'
 #' @return
 #' Returns a list with entries:
@@ -65,31 +80,26 @@
 #' Use the optional \code{center} parameter to implicitly subtract the values
 #' in the \code{center} vector from each column of \code{A}, computing the
 #' truncated SVD of \code{sweep(A, 2, center, FUN=`-`)},
-#' without explicitly forming the centered matrix. This option may not be
-#' used together with the general rank 1 deflation options. \code{center}
+#' without explicitly forming the centered matrix. \code{center}
 #' must be a vector of length equal to the number of columns of \code{A}.
 #' This option may be used to efficiently compute principal components without
 #' explicitly forming the centered matrix (which can, importantly, preserve
 #' sparsity in the matrix). See the examples.
 #'
-#' The optional deflation parameters are deprecated and will be removed in
-#' a future version. They could be used to compute the rank-one deflated
-#' SVD of \eqn{A - ds \cdot du dv^T}{A - ds*du \%*\% t(dv)}, where
-#' \eqn{du^T A - ds\cdot dv^T = 0}{t(du) \%*\% A - ds * t(dv) == 0}. For
-#' example, the triple \code{ds, du, dv} may be a known singular value
-#' and corresponding singular vectors. Or \code{ds=m} and \code{dv}
-#' and \code{du} represent a vector of column means of \code{A} and of ones,
-#' respectively, where \code{m} is the number of rows of \code{A}.
-#' This functionality can be effectively replaced with custom matrix
-#' product functions.
+#' The optional \code{shift} scalar valued argument applies only to square matrices; use it
+#' to estimate the partial svd of \code{A + diag(shift, nrow(A), nrow(A))}
+#' (without explicitly forming the shifted matrix).
 #'
-#' Specify an optional alternative matrix multiplication operator in the
+#' (Deprecated) Specify an optional alternative matrix multiplication operator in the
 #' \code{mult} parameter. \code{mult} must be a function of two arguments,
 #' and must handle both cases where one argument is a vector and the other
-#' a matrix. See the examples.
+#' a matrix. This option is deprecated and will be removed in a future version.
+#' The new preferred method simply uses R itself to define a custom matrix class
+#' with your user-defined matrix multiplication operator. See the examples.
 #'
 #' Use the \code{v} option to supply a starting vector for the iterative
-#' method. A random vector is used by default. Optionally set \code{v} to
+#' method. A random vector is used by default (precede with \code{set.seed()}
+#' for reproducibility). Optionally set \code{v} to
 #' the output of a previous run of \code{irlba} to restart the method, adding
 #' additional singular values/vectors without recomputing the solution
 #' subspace. See the examples.
@@ -111,6 +121,10 @@
 #' The function might return an error for several reasons including a situation when the starting
 #' vector \code{v} is near the null space of the matrix. In that case, try a different \code{v}.
 #'
+#' The \code{fastpath=TRUE} option only supports real-valued matrices and sparse matrices
+#' of type \code{dgCMatrix} (for now). Other problems fall back to the reference
+#' R implementation.
+#'
 #' @references
 #' Augmented Implicitly Restarted Lanczos Bidiagonalization Methods, J. Baglama and L. Reichel, SIAM J. Sci. Comput. 2005.
 #'
@@ -128,6 +142,12 @@
 #' # (starting with an existing solution S)
 #' S1 <- irlba(A, 5, v=S)
 #'
+#' # Estimate smallest singular values
+#' irlba(A, 3, smallest=TRUE)$d
+#'
+#' #Compare with
+#' tail(svd(A)$d, 3)
+#'
 #' # Principal components (see also prcomp_irlba)
 #' P <- irlba(A, nv=1, center=colMeans(A))
 #'
@@ -138,6 +158,7 @@
 #'
 #' # A custom matrix multiplication function that scales the columns of A
 #' # (cf the scale option). This function scales the columns of A to unit norm.
+#' # This approach is deprecated (see below for a bettwe way to do this).
 #' col_scale <- sqrt(apply(A, 2, crossprod))
 #' mult <- function(x, y)
 #'         {
@@ -152,57 +173,77 @@
 #' irlba(A, 3, mult=mult)$d
 #'
 #' # Compare with:
-#' irlba(A, 3, scale=col_scale)$d
-#'
-#' # Compare with:
 #' svd(sweep(A, 2, col_scale, FUN=`/`))$d[1:3]
 #'
-#' @seealso \code{\link{svd}}, \code{\link{prcomp}}, \code{\link{partial_eigen}}
+#' # Compare with the new recommended approach:
+#' setClass("scaled_matrix", contains="matrix", slots=c(scale="numeric"))
+#' setMethod("%*%", signature(x="scaled_matrix", y="numeric"),
+#'    function(x ,y) x@.Data %*% (y / x@scale))
+#' setMethod("%*%", signature(x="numeric", y="scaled_matrix"),
+#'    function(x ,y) (x %*% y@.Data) / y@scale)
+#' a <- new("scaled_matrix", A, scale=col_scale)
+#' irlba(a, 3)$d
+#'
+#' @seealso \code{\link{svd}}, \code{\link{prcomp}}, \code{\link{partial_eigen}}, \code{\link{svdr}}
 #' @import Matrix
 #' @importFrom stats rnorm
 #' @importFrom methods slotNames
-#' @useDynLib irlba
+#' @useDynLib irlba, .registration=TRUE, .fixes="C_"
 #' @export
 irlba <-
-function (A,                     # data matrix
-          nv=5, nu,              # number of singular vectors to estimate
-          maxit=1000,            # maximum number of iterations
-          work=nv + 7,           # working subspace size
-          reorth=TRUE,           # TRUE=full reorthogonalization
-          tol=1e-5,              # stopping tolerance
-          v=NULL,                # optional starting vector or restart
-          right_only=FALSE,      # TRUE=only return V
-          verbose=FALSE,         # display status messages
-          scale,                 # optional column scaling
-          center,                # optional column centering
-          du, ds, dv,            # optional general rank 1 deflation
-          shift,                 # optional shift for square matrices
-          mult,                  # optional custom matrix multiplication func.
-          fastpath=TRUE)         # use the faster C implementation if possible
+function(A,                     # data matrix
+         nv=5, nu=nv,           # number of singular vectors to estimate
+         maxit=100,             # maximum number of iterations
+         work=nv + 7,           # working subspace size
+         reorth=TRUE,           # TRUE=full reorthogonalization
+         tol=1e-5,              # stopping tolerance
+         v=NULL,                # optional starting vector or restart
+         right_only=FALSE,      # TRUE=only return V
+         verbose=FALSE,         # display status messages
+         scale=NULL,            # optional column scaling
+         center=NULL,           # optional column centering
+         shift=NULL,            # optional shift for square matrices
+         mult=NULL,             # optional custom matrix multiplication func.
+         fastpath=TRUE,         # use the faster C implementation if possible
+         svtol=tol,             # stopping tolerance percent change in estimated svs
+         smallest=FALSE,        # set to TRUE to estimate subspaces associated w/smallest singular values
+         ...)                   # optional arguments (really just to support old removed args)
 {
 # ---------------------------------------------------------------------
 # Check input parameters
 # ---------------------------------------------------------------------
   ropts <- options(warn=1) # immediately show warnings
   on.exit(options(ropts))  # reset on exit
+  interchange <- FALSE
   eps <- .Machine$double.eps
-  deflate <- missing(du) + missing(ds) + missing(dv)
+  # hidden support for old, removed (previously deprecated) parameters
+  # this is here as a convenience to keep old code working without change
+  mcall <- as.list(match.call())
+  # Maximum number of Ritz vectors to use in augmentation, may be less
+  # depending on workspace size.
+  maxritz <- mcall[["maxritz"]]
+  if (is.null(maxritz)) maxritz <- 3
+  du <- mcall[["du"]]
+  dv <- mcall[["dv"]]
+  ds <- mcall[["ds"]]
+  deflate <- is.null(du) + is.null(ds) + is.null(dv)
+  if (smallest) fastpath <- FALSE
   if (deflate == 3)
   {
     deflate <- FALSE
   } else if (deflate == 0)
   {
     deflate <- TRUE
-    warning("The deflation options are deprecated and will be removed in a future version.")
+    warning("The deflation options are deprecated and have been removed as formal arugments. Please modify your code to not use them.")
     if (length(ds) > 1) stop("deflation limited to one dimension")
     if (!is.null(dim(du))) du <- du[, 1]
     if (!is.null(dim(dv))) dv <- dv[, 1]
   } else stop("all three du ds dv parameters must be specified for deflation")
-  if (!missing(center))
+  if (!is.null(center))
   {
     if (deflate) stop("the center parameter can't be specified together with deflation parameters")
     if (length(center) != ncol(A)) stop("center must be a vector of length ncol(A)")
-    if (fastpath) du <- NULL
+    if (fastpath && ! right_only) du <- NULL
     else du <- 1
     ds <- 1
     dv <- center
@@ -211,10 +252,10 @@ function (A,                     # data matrix
   iscomplex <- is.complex(A)
   m <- nrow(A)
   n <- ncol(A)
-  if (missing(nu)) nu <- nv
-  if (!missing(mult) && deflate) stop("the mult parameter can't be specified together with deflation parameters")
+  if (is.null(nu)) nu <- nv
+  if (!is.null(mult) && deflate) stop("the mult parameter can't be specified together with deflation parameters")
   missingmult <- FALSE
-  if (missing(mult))
+  if (is.null(mult))
   {
     missingmult <- TRUE
     mult <- `%*%`
@@ -224,7 +265,7 @@ function (A,                     # data matrix
   if (k > min(m - 1, n - 1)) stop("max(nu, nv) must be strictly less than min(nrow(A), ncol(A))")
   if (k >= 0.5 * min(m, n))
   {
-    warning("You're computing a large percentage of total singular values, standard svd might work better!")
+    warning("You're computing too large a percentage of total singular values, use a standard svd instead.")
   }
   if (work <= 1) stop("work must be greater than 1")
   if (tol < 0) stop("tol must be non-negative")
@@ -244,39 +285,57 @@ function (A,                     # data matrix
   if (right_only)
   {
     w_dim <- 1
-    work <- min(min(m, n), work + 10 ) # typically need this to help convergence
+    work <- min(min(m, n), work + 20) # typically need this to help convergence
+    fastpath <- FALSE
+  }
+  if (n > m && smallest)
+  {
+    # Interchange dimensions m,n so that dim(A'A) = min(m,n) when seeking the
+    # smallest singular values; avoids finding zero-valued smallest singular values.
+    interchange <- TRUE
+    temp <- m
+    m <- n
+    n <- temp
   }
 
   if (verbose)
   {
-    message ("Working dimension size ", work)
+    message("Working dimension size ", work)
   }
 # Check for tiny problem, use standard SVD in that case. Make definition of 'tiny' larger?
   if (min(m, n) < 6)
   {
-    if (verbose) warning("Tiny problem detected, using standard `svd` function.")
-    algorithm = "svd"
-    if (!missing(scale)) A <- A / scale
-    if (!missing(shift)) A <- A + diag(shift)
+    if (verbose) message("Tiny problem detected, using standard `svd` function.")
+    if (!is.null(scale)) A <- A / scale
+    if (!is.null(shift)) A <- A + diag(shift, nrow(A), ncol(A))
     if (deflate)
     {
       if (is.null(du)) du <- rep(1, nrow(A))
       A <- A - (ds * du) %*% t(dv)
     }
     s <- svd(A)
+    if (smallest)
+    {
+      return(list(d=tail(s$d, k), u=s$u[, tail(seq(ncol(s$u)), k), drop=FALSE],
+              v=s$v[, tail(seq(ncol(s$v), k)), drop=FALSE], iter=0, mprod=0))
+    }
     return(list(d=s$d[1:k], u=s$u[, 1:nu, drop=FALSE],
               v=s$v[, 1:nv, drop=FALSE], iter=0, mprod=0))
   }
 
 # Try to use the fast C-language code path
   if (deflate) fastpath <- fastpath && is.null(du)
+# Only matrix, dgCMatrix supported by fastpath
+  fastpath <- fastpath && (("Matrix" %in% attributes(class(A)) && ("dgCMatrix" %in% class(A))) || "matrix" %in% class(A))
   if (fastpath && missingmult && !iscomplex && !right_only)
   {
-    RESTART <- 0
+    RESTART <- 0L
     RV <- RW <- RS <- NULL
     if (is.null(v))
+    {
       v <- rnorm(n)
-    else if (is.list(v))  # restarted case
+      if (verbose) message("Initializing starting vector v with samples from standard normal distribution.\nUse `set.seed` first for reproducibility.")
+    } else if (is.list(v))  # restarted case
     {
       if (is.null(v$v) || is.null(v$d) || is.null(v$u)) stop("restart requires left and right singular vectors")
       if (max(nu, nv) <= min(ncol(v$u), ncol(v$v))) return(v) # Nothing to do!
@@ -294,12 +353,12 @@ function (A,                     # data matrix
     SCALE <- NULL
     SHIFT <- NULL
     CENTER <- NULL
-    if (!missing(scale))
+    if (!is.null(scale))
     {
       if (length(scale) != ncol(A)) stop("scale length must mactch number of matrix columns")
       SCALE <- as.double(scale)
     }
-    if (!missing(shift))
+    if (!is.null(shift))
     {
       if (length(shift) != 1) stop("shift length must be 1")
       SHIFT <- as.double(shift)
@@ -309,15 +368,15 @@ function (A,                     # data matrix
       if (length(center) != ncol(A)) stop("the centering vector length must match the number of matrix columns")
       CENTER <- as.double(center)
     }
-    ans <- .Call("IRLB", A, as.integer(k), as.double(v), as.integer(work),
-                 as.integer(maxit), as.double(tol), .Machine$double.eps, as.integer(SP),
-                 RESTART, RV, RW, RS, SCALE, SHIFT, CENTER, PACKAGE="irlba")
+    ans <- .Call(C_IRLB, A, as.integer(k), as.double(v), as.integer(work),
+                 as.integer(maxit), as.double(tol), as.double(.Machine$double.eps), as.integer(SP),
+                 as.integer(RESTART), RV, RW, RS, SCALE, SHIFT, CENTER, as.double(svtol))
     if (ans[[6]] == 0 || ans[[6]] == -2)
     {
       names(ans) <- c("d", "u", "v", "iter", "mprod", "err")
       ans$u <- matrix(head(ans$u, m * nu), nrow=m, ncol=nu)
       ans$v <- matrix(head(ans$v, n * nv), nrow=n, ncol=nv)
-      if(tol * ans$d[1] < eps) warning("convergence criterion below machine epsilon")
+      if (tol * ans$d[1] < eps) warning("convergence criterion below machine epsilon")
       if (ans[[6]] == -2) warning("did not converge--results might be invlaid!; try increasing maxit or fastpath=FALSE")
       return(ans[-6])
     }
@@ -328,7 +387,7 @@ function (A,                     # data matrix
                 "linear dependency encountered")
     erridx <- abs(ans[[6]])
     if (erridx > 1)
-      warning("fast code path error ", errors[erridx], "; re-trying with fastpath=FALSE.")
+      warning("fast code path error ", errors[erridx], "; re-trying with fastpath=FALSE.", immediate.=TRUE)
   }
 
 # Allocate memory for W and F:
@@ -374,7 +433,7 @@ function (A,                     # data matrix
                              # B est. ||A||_2
   Smin <- NULL               # Min value of all computed singular values of
                              # B est. cond(A)
-  SVTol <- tol  # Tolerance for singular vector convergence
+  lastsv <- c()              # estimated sv in last iteration
 
 # Check for user-supplied restart condition
   if (restart)
@@ -415,12 +474,13 @@ function (A,                     # data matrix
 #   Compute W=AV
 #   Optionally apply scale
     VJ <- V[, j]
-    if (!missing(scale))
+    if (!is.null(scale))
     {
       VJ <- VJ / scale
     }
+    if (interchange) avj <- mult(VJ, A)
+    else avj <- mult(A, VJ)
 #   Handle sparse products.
-    avj <- mult(A, VJ)
     if ("Matrix" %in% attributes(class(avj)) && "x" %in% slotNames(avj))
     {
       if (length(avj@x) == nrow(W)) avj <- slot(avj, "x")
@@ -430,7 +490,7 @@ function (A,                     # data matrix
     mprod <- mprod + 1
 
 #   Optionally apply shift
-    if (!missing(shift))
+    if (!is.null(shift))
     {
       W[, j_w] <- W[, j_w] + shift * VJ
     }
@@ -444,14 +504,15 @@ function (A,                     # data matrix
 #   Orthogonalize W
     if (iter != 1 && w_dim > 1 && reorth)
     {
-      W[, j] <- orthog (W[, j, drop=FALSE], W[, 1:(j - 1), drop=FALSE])
+      W[, j] <- orthog(W[, j, drop=FALSE], W[, 1:(j - 1), drop=FALSE])
     }
 
     S <- norm2(W[, j_w, drop=FALSE])
 #   Check for linearly dependent vectors
-    if (S < eps2 && j == 1) stop("starting vector near the null space")
-    if (S < eps2)
+    if (is.na(S) || S < eps2 && j == 1) stop("starting vector near the null space")
+    if (is.na(S) || S < eps2)
     {
+      if (verbose) message("invariant subspace found")
       W[, j_w] <- rnorm(nrow(W))
       if (w_dim > 1) W[, j] <- orthog(W[, j], W[, 1:(j - 1)])
       W[, j_w] <- W[, j_w] / norm2(W[, j_w])
@@ -465,12 +526,17 @@ function (A,                     # data matrix
       j_w <- ifelse(w_dim > 1, j, 1)
       if (iscomplex)
       {
-        F <- Conj(t(drop(mult(Conj(drop(W[, j_w])), A))))
+        if (interchange) F <- Conj(t(drop(mult(A, Conj(drop(W[, j_w]))))))
+        else F <- Conj(t(drop(mult(Conj(drop(W[, j_w])), A))))
+      }
+      else
+      {
+        if (interchange) F <- t(drop(mult(A, drop(W[, j_w]))))
+        else F <- t(drop(mult(drop(W[, j_w]), A)))
       }
-      else F <- t(drop(mult(drop(W[, j_w]), A)))
 #     Optionally apply shift and scale
-      if (!missing(shift)) F <- F + shift * W[, j_w]
-      if (!missing(scale)) F <- F / scale
+      if (!is.null(shift)) F <- F + shift * W[, j_w]
+      if (!is.null(scale)) F <- F / scale
       mprod <- mprod + 1
       F <- drop(F - S * V[, j])
 #     Orthogonalize
@@ -481,6 +547,7 @@ function (A,                     # data matrix
 #       Check for linear dependence
         if (R < eps2)
         {
+          if (verbose) message("invariant subspace found")
           F <- matrix(rnorm(dim(V)[1]), dim(V)[1], 1)
           F <- orthog(F, V[, 1:j, drop=FALSE])
           V[, j + 1] <- F / norm2(F)
@@ -497,15 +564,16 @@ function (A,                     # data matrix
 
 #       Optionally apply scale
         VJP1 <- V[, j + 1]
-        if (!missing(scale))
+        if (!is.null(scale))
         {
           VJP1 <- VJP1 / scale
         }
-        W[, jp1_w] <- drop(mult(A, drop(VJP1)))
+        if (interchange) W[, jp1_w] <- drop(mult(drop(VJP1), A))
+        else W[, jp1_w] <- drop(mult(A, drop(VJP1)))
         mprod <- mprod + 1
 
 #       Optionally apply shift
-        if (!missing(shift))
+        if (!is.null(shift))
         {
           W[, jp1_w] <- W[, jp1_w] + shift * VJP1
         }
@@ -525,6 +593,7 @@ function (A,                     # data matrix
 #       Check for linear dependence
         if (S < eps2)
         {
+          if (verbose) message("invariant subspace found")
           W[, jp1_w] <- rnorm(nrow(W))
           if (w_dim > 1) W[, j + 1] <- orthog(W[, j + 1], W[, 1:j])
           W[, jp1_w] <- W[, jp1_w] / norm2(W[, jp1_w])
@@ -539,18 +608,14 @@ function (A,                     # data matrix
       }
       j <- j + 1
     }
-    if (verbose)
-    {
-      message("Lanczos iter = ", iter, ", dim = ", j - 1, ", mprod = ", mprod)
-    }
 # ---------------------------------------------------------------------
 # (End of the Lanczos bidiagonalization part)
 # ---------------------------------------------------------------------
     Bsz <- nrow(B)
     R_F <- norm2(F)
     F <- F / R_F
-#   Compute singular triplets of B. Expect svd to return s.v.s in order
-#   from largest to smallest.
+#   Compute singular triplets of B, svd must return ordered singular
+#   values from largest to smallest.
     Bsvd <- svd(B)
 
 #   Estimate ||A|| using the largest singular value over all iterations
@@ -568,28 +633,71 @@ function (A,                     # data matrix
       Smin <- min(Smin, Bsvd$d[Bsz])
     }
     Smax <- max(eps23, Smax)
-    if (Smin / Smax < sqrteps && !reorth)
+    if (! reorth && Smin / Smax < sqrteps)
     {
       warning("The matrix is ill-conditioned. Basis will be reorthogonalized.")
       reorth <- TRUE
     }
+    if (smallest)
+    {
+      jj <- seq(ncol(Bsvd$u), 1, by = -1)
+      Bsvd$u <- Bsvd$u[, jj]
+      Bsvd$d <- Bsvd$d[jj]
+      Bsvd$v <- Bsvd$v[, jj]
+    }
 
 #   Compute the residuals
     R <- R_F * Bsvd$u[Bsz, , drop=FALSE]
 #   Check for convergence
-    ct <- convtests(Bsz, tol, k_org, Bsvd$u,
-                    Bsvd$d, Bsvd$v, abs(R), k, SVTol, Smax)
+    ct <- convtests(Bsz, tol, k_org, Bsvd, abs(R), k, Smax, lastsv, svtol, maxritz, work, S)
+    if (verbose)
+    {
+      message("iter= ", iter,
+              ", mprod= ", mprod,
+              ", sv[", k_org, "]=", sprintf("%.2e", Bsvd$d[k_org]),
+              ", %change=", sprintf("%.2e", (Bsvd$d[k_org] - lastsv[k_org])/Bsvd$d[k_org]),
+              ", k=", ct$k)
+    }
+    lastsv <- Bsvd$d
     k <- ct$k
 
 #   If all desired singular values converged, then exit main loop
     if (ct$converged) break
     if (iter >= maxit) break
 
-#   Compute the starting vectors and first block of B[1:k, 1:(k+1), drop=FALSE]
+#   Compute the starting vectors and first block of B
+    if (smallest && (Smin / Smax > sqrteps))
+    {
+#     Update the SVD of B to be the svd of [B ||F||E_m]
+      Bsvd2.d <- Bsvd$d
+      Bsvd2.d <- diag(Bsvd2.d, nrow=length(Bsvd2.d))
+      Bsvd2 <- svd(cbind(Bsvd2.d, t(R)))
+      jj <- seq(ncol(Bsvd2$u), 1, by=-1)
+      Bsvd2$u <- Bsvd2$u[, jj]
+      Bsvd2$d <- Bsvd2$d[jj]
+      Bsvd2$v <- Bsvd2$v[, jj]
+
+      Bsvd$d <- Bsvd2$d
+      Bsvd$u <- Bsvd$u %*% Bsvd2$u
+      Bsvd$v <- cbind(rbind(Bsvd$v, rep(0, Bsz)), c(rep(0, Bsz), 1)) %*% Bsvd2$v
+      V_B_last <- Bsvd$v[Bsz + 1, 1:k]
+      s <- R_F * solve(B, cbind(c(rep(0, Bsz - 1), 1)))
+      Bsvd$v <- Bsvd$v[1:Bsz, , drop=FALSE] + s %*% Bsvd$v[Bsz + 1, ]
+
+      qrv <- qr(cbind(rbind(Bsvd$v[, 1:k], 0), rbind(-s, 1)))
+      Bsvd$v <- qr.Q(qrv)
+      R <- qr.R(qrv)
+      V[, 1:(k + 1)] <- cbind(V, F) %*% Bsvd$v
+
+#  Update and compute the k by k+1 part of B
+      UT <- t(R[1:(k + 1), 1:k] + R[, k + 1] %*% rbind(V_B_last))
+      B <- diag(Bsvd$d[1:k], nrow=k) %*% (UT * upper.tri(UT, diag=TRUE))[1:k, 1:(k+1)]
+    } else
+    {
 #   using the Ritz vectors
-      V[, 1:(k + dim(F)[2])] <- cbind(V[, 1:(dim(Bsvd$v)[1]),
-                                     drop=FALSE] %*% Bsvd$v[, 1:k], F)
-      B <- cbind( diag(Bsvd$d[1:k], nrow=k), R[1:k])
+      V[, 1:(k + dim(F)[2])] <- cbind(V[, 1:(dim(Bsvd$v)[1]), drop=FALSE] %*% Bsvd$v[, 1:k], F)
+      B <- cbind(diag(Bsvd$d[1:k], nrow=k), R[1:k])
+    }
 
 #   Update the left approximate singular vectors
     if (w_dim > 1)
@@ -603,14 +711,21 @@ function (A,                     # data matrix
 # End of the main iteration loop
 # Output results
 # ---------------------------------------------------------------------
-  if (iter > maxit) warning("did not converge--results might be invalid!; try increasing maxit")
+  if (!ct$converged) warning("did not converge--results might be invalid!; try increasing maxit or work")
   d <- Bsvd$d[1:k_org]
   if (!right_only)
   {
     u <- W[, 1:(dim(Bsvd$u)[1]), drop=FALSE] %*% Bsvd$u[, 1:k_org, drop=FALSE]
   }
   v <- V[, 1:(dim(Bsvd$v)[1]), drop=FALSE] %*% Bsvd$v[, 1:k_org, drop=FALSE]
-  if(tol * d[1] < eps) warning("convergence criterion below machine epsilon")
+  if (smallest)
+  {
+    reverse <- seq(length(d), 1)
+    d <- d[reverse]
+    if (!right_only) u <- u[, reverse]
+    v <- v[, reverse]
+  }
+  if (tol * d[1] < eps) warning("convergence criterion below machine epsilon")
   if (right_only)
     return(list(d=d, v=v[, 1:nv, drop=FALSE], iter=iter, mprod=mprod))
   return(list(d=d, u=u[, 1:nu, drop=FALSE],
